@@ -27,6 +27,7 @@ import errno
 import struct
 import logging
 import traceback
+import random
 import encrypt
 import eventloop
 from common import parse_header
@@ -96,6 +97,8 @@ class TCPRelayHandler(object):
         self._upstream_status = WAIT_STATUS_READING
         self._downstream_status = WAIT_STATUS_INIT
         self._remote_address = None
+        if is_local:
+            self._chosen_server = self._get_a_server()
         fd_to_handlers[local_sock.fileno()] = self
         local_sock.setblocking(False)
         local_sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
@@ -112,6 +115,15 @@ class TCPRelayHandler(object):
     def remote_address(self):
         return self._remote_address
 
+    def _get_a_server(self):
+        server = self._config['server']
+        server_port = self._config['server_port']
+        if type(server_port) == list:
+            server_port = random.choice(server_port)
+        logging.debug('chosen server: %s:%d', server, server_port)
+        # TODO support multiple server IP
+        return server, server_port
+
     def _update_activity(self):
         self._server.update_activity(self)
 
@@ -157,7 +169,8 @@ class TCPRelayHandler(object):
                 uncomplete = True
             else:
                 logging.error(e)
-                traceback.print_exc()
+                if self._config['verbose']:
+                    traceback.print_exc()
                 self.destroy()
                 return False
         if uncomplete:
@@ -189,8 +202,7 @@ class TCPRelayHandler(object):
                 data = ''.join(self._data_to_write_to_local)
                 l = len(data)
                 s = self._remote_sock.sendto(data, MSG_FASTOPEN,
-                                             (self._config['server'],
-                                              self._config['server_port']))
+                                             self._chosen_server)
                 if s < l:
                     data = data[s:]
                     self._data_to_write_to_local = [data]
@@ -208,7 +220,8 @@ class TCPRelayHandler(object):
                     self.destroy()
                 else:
                     logging.error(e)
-                    traceback.print_exc()
+                    if self._config['verbose']:
+                        traceback.print_exc()
                     self.destroy()
 
     def _handle_stage_hello(self, data):
@@ -253,7 +266,7 @@ class TCPRelayHandler(object):
                 data_to_send = self._encryptor.encrypt(data)
                 self._data_to_write_to_remote.append(data_to_send)
                 # notice here may go into _handle_dns_resolved directly
-                self._dns_resolver.resolve(self._config['server'],
+                self._dns_resolver.resolve(self._chosen_server[0],
                                            self._handle_dns_resolved)
             else:
                 if len(data) > header_length:
@@ -263,7 +276,8 @@ class TCPRelayHandler(object):
                                            self._handle_dns_resolved)
         except Exception as e:
             logging.error(e)
-            traceback.print_exc()
+            if self._config['verbose']:
+                traceback.print_exc()
             # TODO use logging when debug completed
             self.destroy()
 
@@ -280,8 +294,7 @@ class TCPRelayHandler(object):
                     remote_addr = self._remote_address[0]
                     remote_port = self._remote_address[1]
                     if self._is_local:
-                        remote_addr = self._config['server']
-                        remote_port = self._config['server_port']
+                        remote_addr, remote_port = self._chosen_server
                     addrs = socket.getaddrinfo(ip, remote_port, 0,
                                                socket.SOCK_STREAM,
                                                socket.SOL_TCP)
@@ -317,7 +330,8 @@ class TCPRelayHandler(object):
                     return
                 except (OSError, IOError) as e:
                     logging.error(e)
-                    traceback.print_exc()
+                    if self._config['verbose']:
+                        traceback.print_exc()
         self.destroy()
 
     def _on_local_read(self):
@@ -375,7 +389,8 @@ class TCPRelayHandler(object):
             self._write_to_sock(data, self._local_sock)
         except Exception as e:
             logging.error(e)
-            traceback.print_exc()
+            if self._config['verbose']:
+                traceback.print_exc()
             # TODO use logging when debug completed
             self.destroy()
 
@@ -397,19 +412,20 @@ class TCPRelayHandler(object):
             self._update_stream(STREAM_UP, WAIT_STATUS_READING)
 
     def _on_local_error(self):
+        logging.debug('got local error')
         if self._local_sock:
-            logging.debug('got local error')
             logging.error(eventloop.get_sock_error(self._local_sock))
         self.destroy()
 
     def _on_remote_error(self):
+        logging.debug('got remote error')
         if self._remote_sock:
-            logging.debug('got remote error')
             logging.error(eventloop.get_sock_error(self._remote_sock))
         self.destroy()
 
     def handle_event(self, sock, event):
         if self._stage == STAGE_DESTROYED:
+            logging.debug('ignore handle_event: destroyed')
             return
         # order is important
         if sock == self._remote_sock:
@@ -439,6 +455,7 @@ class TCPRelayHandler(object):
 
     def destroy(self):
         if self._stage == STAGE_DESTROYED:
+            logging.debug('already destroyed')
             return
         self._stage = STAGE_DESTROYED
         if self._remote_address:
@@ -447,11 +464,13 @@ class TCPRelayHandler(object):
         else:
             logging.debug('destroy')
         if self._remote_sock:
+            logging.debug('destroying remote')
             self._loop.remove(self._remote_sock)
             del self._fd_to_handlers[self._remote_sock.fileno()]
             self._remote_sock.close()
             self._remote_sock = None
         if self._local_sock:
+            logging.debug('destroying local')
             self._loop.remove(self._local_sock)
             del self._fd_to_handlers[self._local_sock.fileno()]
             self._local_sock.close()
@@ -539,6 +558,7 @@ class TCPRelay(object):
         # we just need a sorted last_activity queue and it's faster than heapq
         # in fact we can do O(1) insertion/remove so we invent our own
         if self._timeouts:
+            logging.debug('sweeping timeouts')
             now = time.time()
             length = len(self._timeouts)
             pos = self._timeout_offset
@@ -569,15 +589,15 @@ class TCPRelay(object):
 
     def _handle_events(self, events):
         for sock, fd, event in events:
-            if sock:
-                logging.debug('fd %d %s', fd,
-                              eventloop.EVENT_NAMES.get(event, event))
+            # if sock:
+            #     logging.debug('fd %d %s', fd,
+            #                   eventloop.EVENT_NAMES.get(event, event))
             if sock == self._server_socket:
                 if event & eventloop.POLL_ERR:
                     # TODO
                     raise Exception('server_socket error')
                 try:
-                    # logging.debug('accept')
+                    logging.debug('accept')
                     conn = self._server_socket.accept()
                     TCPRelayHandler(self, self._fd_to_handlers, self._eventloop,
                                     conn[0], self._config, self._dns_resolver,
@@ -588,7 +608,8 @@ class TCPRelay(object):
                         continue
                     else:
                         logging.error(e)
-                        traceback.print_exc()
+                        if self._config['verbose']:
+                            traceback.print_exc()
             else:
                 if sock:
                     handler = self._fd_to_handlers.get(fd, None)
